#ifndef PANGULU_COMMON_H
#define PANGULU_COMMON_H
/*
    An impoortant structure is structure pangulu_Smatrix
    {
        calculate_type *value;
        calculate_type *value_CSC;
        int_t *CSR_to_CSC_index;
        int_t *CSC_to_CSR_index;
        int_t *rowpointer;
        int_t *columnindex;
        int_t *columnpointer;
        int_t *rowindex;
        int_t column;
        int_t row;
        int_t nnz;
        int_32t *nnzU;
        int_t *bin_rowpointer;
        int_t *bin_rowindex;
        int_t *CUDA_rowpointer;
        int_t *CUDA_columnindex;
        calculate_type *CUDA_value;
        int_32t *CUDA_nnzU;
        int_t *CUDA_bin_rowpointer;
        int_t *CUDA_bin_rowindex;
    } pangulu_Smatrix;
It's parameters are explained as follows: 
value is the value stored in CSR format;
value_CSC is the value stored in CSC format;
CSR_to_CSC_index is CSR converted to CSC storage index;
CSC_to_CSR_index is CSC converted to CSR storage index;
rowpointer is the rowpointer stored in CSR format;
columnindex is the columnindex stored in CSR format;
columnrowpointer is the columnpointer stored in CSC format;
rowindex is the rowindex stored in CSC format;
row is this matrix row;
column is this matrix column;
nnz is the number of non-zero in this matrix;
bin_rowpointer is The matrix csr is adjusted and decomposed into different intervals 
according to the number of non-zero elements in each row. According to 1-32, 33-64, 
65-128, 129-256, 257-512, 513-1024, 1025-2048, 2049-4096, 4097-infinity.
the distribution is performed 
bin_rowindex Save the row index of bin_rowpointer. 
nnzU saves the number of non-zero elements in each row of the matrix .
CUDA_value is the value stored in CSR format and stored in CUDA_memory;
CUDA_nnzU is the nnzU stored in CUDA_memory;
CUDA_bin_rowpointer is bin_rowpointer stored in CUDA_memory;
CUDA_bin_rowindex is bin_rowindex stored in CUDA_memory;
CUDA_rowpointer is the rowpointer stored in CSR format and stored in CUDA_memory;
CUDA_columnindex is the columnindex stored in CSR format and stored in CUDA_memory;
    
    An impoortant structure is structure pangulu_vector
    {
        calculate_type *value;
        int_t row;
    } pangulu_vector;
It's parameters are explained as follows: 
value is the value stored in dense vector;
row is the vector stored the vector's latitude;

    An impoortant structure is structure pangulu_common
    {
        int_32t P;
        int_32t Q;
        int_32t rank;
        int_t N;
        int_32t NB;
        char *file_name;
        int_32t sum_rank_size;
        struct timeval start_time;
        struct timeval stop_time;
    } pangulu_common;
It's parameters are explained as follows: 
P is number of row nodes;
Q is number of col nodes;
rank is process ID of the current node;
N is the matrix size;
NB is the matrix block size;
file_name is the rank 0 read the matrix name;
sum_rank_size is all processes number;
start_time is the test begin time;
stop_time is the test end time;

    An impoortant structure is structure pangulu_common
    {
        int_32t P;
        int_32t Q;
        int_32t rank;
        int_t N;
        int_32t NB;
        char *file_name;
        int_32t sum_rank_size;
        struct timeval start_time;
        struct timeval stop_time;
    } pangulu_common;
It's parameters are explained as follows: 
P is number of row nodes;
Q is number of col nodes;
rank is process ID of the current node;
N is the matrix size;
NB is the matrix block size;
file_name is the rank 0 read the matrix name;
sum_rank_size is all processes number;
start_time is the test begin time;
stop_time is the test end time;

    An impoortant structure is structure pangulu_block_Smatrix to store panguLU runtime usage memory;
    An impoortant structure is structure pangulu_block_common to store panguLU runtime usage parameters ;
      
*/
#define CHECK_TIME
#define OVERLAP
#define GPU_TSTRF//define TSTRF kernel calculates by GPU
#define GPU_GESSM//define GESSM kernel calculates by GPU

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
    

typedef int64_t int_t;
typedef int32_t int_32t;
#define MPI_INT_TYPE MPI_LONG

#define calculate_type double
// #if calculate_type==double
         #define MPI_VAL_TYPE MPI_DOUBLE
// #elif calculate_type==float
         //#define MPI_VAL_TYPE MPI_FLOAT
// #endif

#define ERROR 0.000001
#define CHECK_ERROR 0.0001
#define BIN_LENGTH 12
#define pangulu_exchange_PQ(row, P) \
    (row + P - 1) % P

#define calculate_offset(offset_init,now_level,PQ_length) \
((offset_init-now_level%PQ_length)<0)?(offset_init-now_level%PQ_length+PQ_length):(offset_init-now_level%PQ_length)

#define pangulu_Calculate_Block(nrow, block) \
    nrow / block + ((nrow % block) ? 1 : 0)
    
extern int_t CPU_MEMORY;
extern int_t GPU_MEMORY;

#ifdef CHECK_TIME

extern double TIME_transport ;
extern double TIME_isend ;
extern double TIME_receive ;
extern double TIME_getrf ;
extern double TIME_tstrf ;
extern double TIME_gessm ;
extern double TIME_ssssm ;
extern struct timeval GET_TIME_START, GET_TIME_END;
#endif

extern int_32t RANK;
extern int_32t LEVEL;
    typedef struct compare_struct{
        int_t row;
        int_t col;
        int_t kernel_id;
        int_t task_level;
        int_t compare_flag;
    } compare_struct;

#ifdef OVERLAP
    
    typedef struct bsem {
        pthread_mutex_t *mutex;
        pthread_cond_t *cond;
        int_t v;
    } bsem;

#endif

    typedef struct pangulu_heap{
        int_t length;
        int_t *heap_queue;
        compare_struct *comapre_queue;
        int_t max_length;
        int_t nnz_flag;

#ifdef OVERLAP
        bsem *heap_bsem;
#endif

    } pangulu_heap;




    typedef struct pangulu_Smatrix
    {
        calculate_type *value;
        calculate_type *value_CSC;
        int_t *CSR_to_CSC_index;
        int_t *CSC_to_CSR_index;
        int_t *rowpointer;
        int_t *columnindex;
        int_t *columnpointer;
        int_t *rowindex;
        int_t column;
        int_t row;
        int_t nnz;
        int_32t *nnzU;
        int_t *bin_rowpointer;
        int_t *bin_rowindex;
        int_t *CUDA_rowpointer;
        int_t *CUDA_columnindex;
        calculate_type *CUDA_value;
        int_32t *CUDA_nnzU;
        int_t *CUDA_bin_rowpointer;
        int_t *CUDA_bin_rowindex;
    } pangulu_Smatrix;

    typedef struct pangulu_vector
    {
        calculate_type *value;
        int_t row;
    } pangulu_vector;
    

    typedef struct pangulu_common
    {
        int_32t P;
        int_32t Q;
        int_32t rank;
        int_t N;
        int_32t NB;
        int_32t size;
        char *file_name;
        int_32t sum_rank_size;
        struct timeval start_time;
        struct timeval stop_time;
    } pangulu_common;

    typedef struct pangulu_block_Smatrix{
        int_t *block_Smatrix_nnzA_num;
        int_t *block_Smatrix_non_zero_vector_L;
        int_t *block_Smatrix_non_zero_vector_U;
        int_t *mapper_Big_pangulu_Smatrix;
        int_t *Big_pangulu_Smatrix_rowpointer;
        int_t *Big_pangulu_Smatrix_columnindex;
        pangulu_Smatrix **Big_pangulu_Smatrix_value;
        pangulu_Smatrix **Big_pangulu_Smatrix_copy_value;
        int_t *L_pangulu_Smatrix_columnpointer;
        int_t *L_pangulu_Smatrix_rowindex;
        int_t L_Smatrix_nzz;
        pangulu_Smatrix **L_pangulu_Smatrix_value;
        int_t *U_pangulu_Smatrix_rowpointer;
        int_t *U_pangulu_Smatrix_columnindex;
        int_t U_Smatrix_nzz;
        pangulu_Smatrix **U_pangulu_Smatrix_value;
        int_t *mapper_diagonal;
        pangulu_Smatrix **diagonal_Smatrix_L;
        pangulu_Smatrix **diagonal_Smatrix_U;
        pangulu_Smatrix *calculate_L;
        pangulu_Smatrix *calculate_U;
        pangulu_Smatrix *calculate_X;
        
        int_t task_level_length;
        int_t *task_level_num;
        int_t *mapper_LU;
        int_t *task_flag_id;
        pangulu_heap *heap;
        int_t *now_level_L_length;
        int_t *now_level_U_length;
        int_t *save_now_level_L;
        int_t *save_now_level_U;
        int_t *send_flag;
        int_t *send_diagonal_flag_L;
        int_t *send_diagonal_flag_U;
        int_t *grid_process_id;
        int_t *save_send_rank_flag;
        int_t *level_task_rank_id;
        int_t *real_matrix_flag;
        int_t *sum_flag_block_num;
        int_t *receive_level_num;

#ifdef OVERLAP
        bsem *run_bsem1;
        bsem *run_bsem2;
        pthread_mutex_t *MPI_mutex;

#endif


    } pangulu_block_Smatrix;

    typedef struct pangulu_block_common{
        int_t N;
        int_32t rank;
        int_32t P;
        int_32t Q;
        int_32t NB;
        int_32t block_length;
        int_32t sum_rank_size;
        int_32t max_PQ;
        int_32t every_level_length;
    } pangulu_block_common;

#ifdef OVERLAP

    typedef struct thread_param{
        pangulu_block_common *common;
        pangulu_block_Smatrix *Smatrix;
    } thread_param;

#endif


#endif