#ifndef PANGULU_KERNEL_INTERFACE_H
#define PANGULU_KERNEL_INTERFACE_H
/*
The function in this file is primarily called kernel,these functions would be called in pangulu_numerical.h
*/
#include "pangulu_common.h"

#ifdef CHECK_TIME
#include "pangulu_time.h"
#endif


#include "pangulu_getrf_fp64_cuda.h"
#include "pangulu_tstrf_fp64.h"
#include "pangulu_tstrf_fp64_cuda.h"
#include "pangulu_gessm_fp64.h"
#include "pangulu_gessm_fp64_cuda.h"
#include "pangulu_ssssm_fp64_cuda.h"

#include "pangulu_addmatrix_cuda.h"

    void pangulu_getrf_interface(pangulu_Smatrix *A,
                                 pangulu_Smatrix *L,
                                 pangulu_Smatrix *U)
    {

        //pangulu_cuda_memcpy_value_device_to_host_CSC(A);
        //pangulu_display_pangulu_Smatrix_CSC(A);

#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        pangulu_time_check_begin();
#endif

        pangulu_getrf_fp64_cuda(A, L, U);
        
#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        TIME_getrf+=pangulu_time_check_end();
#endif

    }

void pangulu_gessm_interface(pangulu_Smatrix *A,
                             pangulu_Smatrix *L,
                             pangulu_Smatrix *X)
    {
#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        pangulu_time_check_begin();
#endif
        pangulu_gessm_fp64_cuda(A, L, X);
        

#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        TIME_gessm+=pangulu_time_check_end();
#endif
    }
    
    void pangulu_gessm_interface_CPU(pangulu_Smatrix *A,
                                     pangulu_Smatrix *L,
                                     pangulu_Smatrix *X)
    {
#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        pangulu_time_check_begin();
#endif
        pangulu_gessm_fp64(A, L, X);
        

#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        TIME_gessm+=pangulu_time_check_end();
#endif
    }

    void pangulu_tstrf_interface(pangulu_Smatrix *A,
                                 pangulu_Smatrix *X,
                                 pangulu_Smatrix *U)
    {
#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        pangulu_time_check_begin();
#endif

        pangulu_tstrf_fp64_cuda(A, X, U);
        

#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        TIME_tstrf+=pangulu_time_check_end();
#endif

    }

    void pangulu_tstrf_interface_CPU(pangulu_Smatrix *A,
                                     pangulu_Smatrix *X,
                                     pangulu_Smatrix *U)
    {
#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        pangulu_time_check_begin();
#endif

        pangulu_tstrf_fp64(A, X, U);
        

#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        TIME_tstrf+=pangulu_time_check_end();
#endif

    }


void pangulu_ssssm_interface(pangulu_Smatrix *A,
                             pangulu_Smatrix *L,
                             pangulu_Smatrix *U)
    {

#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        pangulu_time_check_begin();
#endif

        pangulu_ssssm_fp64_cuda(A, L, U);
        
#ifdef CHECK_TIME
        pangulu_cuda_devicesynchronize();
        TIME_ssssm+=pangulu_time_check_end();
#endif
    }

void pangulu_addmatrix_interface(pangulu_Smatrix *A,
                                      pangulu_Smatrix *B)
    {
        pangulu_add_pangulu_Smatrix_cuda(A,B);
    }

void pangulu_addmatrix_interface_CPU(pangulu_Smatrix *A,
                                 pangulu_Smatrix *B)
    {
        pangulu_add_pangulu_Smatrix_cpu(A,B);
    }





#endif