#ifndef PANGULU_UTILS_H
#define PANGULU_UTILS_H
/*
The function in this file is primarily called kernel,these functions would be called in many other head files
*/
#include "mmio.h"
#include "mmio_highlevel.h"
#include "pangulu_common.h"

#ifdef CHECK_TIME
#include "pangulu_time.h"
#endif


#include <string.h>
#include "pangulu_mpi.h"
#include "pangulu_malloc.h"
#include "pangulu_cuda_interface.h"
#include <getopt.h>


#define PANGULU_MAX(A,B)        \
    (((A)>(B))?(A):(B))
#define PANGULU_MIN(A,B)        \
    (((A)<(B))?(A):(B))
#define PANGULU_ABS(A)          \
    ((A>0)?(A):(-A))

    void pangulu_get_common(pangulu_common *common,
                            int_t ARGC, char **ARGV)
    {
        int_t c;
        extern char *optarg;
        common->NB = 0;
        common->P = 0;
        common->Q = 0;
        common->sum_rank_size = 0;
        while ((c = getopt(ARGC, ARGV, "P:Q:NB:R:F:")) != EOF)
        {
            switch (c)
            {
            case 'P':
                common->P = atoi(optarg);
                continue;
            case 'Q':
                common->Q = atoi(optarg);
                continue;
            case 'B':
                common->NB = atoi(optarg);
                continue;
            case 'F':
                common->file_name = (char *)pangulu_malloc((strlen(optarg)+10) * sizeof(char));
                strcpy(common->file_name, optarg);
                continue;
            }
        }
        if (common->rank == 0)
        {
            printf("rank is %d P is %d Q is %d NB is %d\n", common->rank, common->P, common->Q, common->NB);
            printf("test matrix name is %s\n", common->file_name);
            fflush(NULL);
        }
        common->sum_rank_size=common->P * common->Q;
        if ((common->NB) == 0)
        {
            printf("error the NB is 0\n");
            exit(4);
        }
    }


    void pangulu_init_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        S->value = NULL;
        S->value_CSC = NULL;
        S->CSR_to_CSC_index = NULL;
        S->CSC_to_CSR_index = NULL;
        S->rowpointer = NULL;
        S->columnindex = NULL;
        S->columnpointer = NULL;
        S->rowindex = NULL;
        S->column = 0;
        S->row = 0;
        S->nnz = 0;

        S->nnzU=NULL;
        S->bin_rowpointer=NULL;
        S->bin_rowindex=NULL;
        S->CUDA_rowpointer=NULL;
        S->CUDA_columnindex=NULL;
        S->CUDA_value=NULL;
        S->CUDA_nnzU=NULL;
        S->CUDA_bin_rowpointer=NULL;
        S->CUDA_bin_rowindex=NULL;

    }

    void pangulu_binary_write_csv_pangulu_Smatrix(pangulu_Smatrix *S,char *filename)
    {
        int_t nnz=S->nnz;
        int_t nrow=S->row;
        int_t ncol=S->column;
        /*
        char copy_filename[256];
        strcpy(copy_filename,filename);
        copy_filename[strlen(filename)-1]='v';
        copy_filename[strlen(filename)-2]='s';
        copy_filename[strlen(filename)-3]='t';
        
        */
        printf("write binary file %s\n", filename);
        int_t *rowpointer = S->rowpointer;
        int_t *columnindex = S->columnindex;
        calculate_type *value = S->value;
        FILE *fp = fopen(filename, "w+");
        if (fp != NULL)
        {
            printf("write nrow is %ld ncol is %ld nnz is %ld\n", nrow, ncol, nnz);
            fwrite(&nrow, sizeof(int_t), 1, fp);
            fwrite(&ncol, sizeof(int_t), 1, fp);
            fwrite(&nnz, sizeof(int_t), 1, fp);
            fwrite(rowpointer, sizeof(int_t), nrow + 1, fp);
            fwrite(columnindex, sizeof(int_t), nnz, fp);
            fwrite(value, sizeof(calculate_type), nnz, fp);
        }
        else
        {
            printf("open error\n");
            return;
        }

        fclose(fp);
    }
    
        int_t pangulu_binary_read_csv_pangulu_Smatrix(pangulu_Smatrix *S,char *filename)
    {   
        int_t nrow, ncol;
        int_t nnz;
        FILE *fp = NULL;
        fp = fopen(filename,"r");
    	if(fp == NULL)
        {
            printf("binary file %s don't exit\n",filename);
            return -1;
        }
        if(1!=fread(&nrow,sizeof(int_t),1,fp)){
            printf("read nrow error\n");
            return -1;
        }
        
        if(1!=fread(&ncol,sizeof(int_t),1,fp)){
            printf("read ncol error\n");
            return -1;
        }
        if(1!=fread(&nnz,sizeof(int_t),1,fp)){
            printf("read nnz error\n");
            return -1;
        }
        int_t *rowpointer = (int_t *)pangulu_malloc(sizeof(int_t) * (nrow + 1));
        if((unsigned int_t)(nrow+1)!=fread(rowpointer,sizeof(int_t),nrow+1,fp)){
            printf("read rowptr error\n");
            return -1;
        }
        //printf("nnz = %ld\n",nnz);
        int_t *columnindex = (int_t *)pangulu_malloc(sizeof(int_t) * nnz);
        calculate_type *value=(calculate_type *)pangulu_malloc(sizeof(calculate_type)*nnz);
        
        if((unsigned int_t)nnz!=fread(columnindex,sizeof(int_t),nnz,fp)){
            printf("read columnindex error\n");
            return -1;
        }
        if((unsigned int_t)nnz!=fread(value,sizeof(calculate_type),nnz,fp)){
            printf("read value error\n");
            return -1;
        }

        S->row=nrow;
        S->column=ncol;
        S->rowpointer=rowpointer;
        S->columnindex=columnindex;
        S->value=value;
        S->nnz=nnz;
        fclose(fp);
        
        return 0;

    }

    void pangulu_read_pangulu_Smatrix(pangulu_Smatrix *S,
                                      char *filename)
    {
        int_t isSymmeticeR;
        int_t nnz;
        int_t nrow, ncol;
        char binary_name[256];
        int_t file_length=strlen(filename);
        strcpy(binary_name,filename);
        binary_name[file_length-1]='v';
        binary_name[file_length-2]='s';
        binary_name[file_length-3]='t';
        int_t flag=pangulu_binary_read_csv_pangulu_Smatrix(S,binary_name);
        if(flag==-1){
            printf("read file name is %s\n",filename);
            mmio_info(&nrow, &ncol, &nnz, &isSymmeticeR, filename);
            int_t *rowpointer = (int_t *)pangulu_malloc(sizeof(int_t) * (nrow + 1));
            int_t *columnindex = (int_t *)pangulu_malloc(sizeof(int_t) * nnz);
            VALUE_TYPE *save_value = (VALUE_TYPE *)pangulu_malloc(sizeof(VALUE_TYPE) * nnz);
            mmio_data(rowpointer, columnindex, save_value, filename);
            S->row = nrow;
            S->column = ncol;
            S->rowpointer = rowpointer;
            S->columnindex = columnindex;
            calculate_type *value = (calculate_type *)pangulu_malloc(sizeof(calculate_type) * nnz);
            for (int_t i = 0; i < nnz; i++)
            {
                value[i] = (calculate_type)(save_value[i]);
            }
            free(save_value);
            S->nnz=nnz;
            S->value = value;
            printf("begin write binary file\n");
            pangulu_binary_write_csv_pangulu_Smatrix(S,binary_name);
            printf("end write binary file\n");
        }
        else{
            printf("read binary file %s\n",binary_name);
        }
        //pangulu_display_pangulu_Smatrix(S);
    }

    void pangulu_write_pangulu_Smatrix(pangulu_Smatrix *S, char *string)
    {

        FILE *fp = fopen(string, "w+");

        if (fp == NULL)
        {
            printf("eeror the file don't open\n");
            return;
        }
        int_t column = S->column;
        int_t row = S->row;
        fprintf(fp, "%%%%MatrixMarket matrix coordinate real general\n");
        fprintf(fp, "%ld ", row);
        fprintf(fp, "%ld ", column);
        fprintf(fp, "%ld\n", S->rowpointer[row]);

        for (int_t i = 0; i < row; i++)
        {
            for (int_t j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                fprintf(fp, "%ld %ld %f\n", S->columnindex[j] + 1, i + 1, S->value[j]);
            }
        }

        fclose(fp);
    }



    void pangulu_time_start(pangulu_common *common)
    {
        gettimeofday(&(common->start_time), NULL);
    }
    
    void pangulu_time_stop(pangulu_common *common)
    {
        gettimeofday(&(common->stop_time), NULL);
    }

    

    int_t pangulu_mapper_A_Smatrix(int_t row, int_t col, int_t *mapper_A, int_t col_length, int_t P, int_t Q)
    {
        int_t mapper_index = mapper_A[(row / P) * col_length + (col / Q)];
        return mapper_index;
    }

    void pangulu_display_pangulu_Smatrix(pangulu_Smatrix *S)
    {
        printf("------------------\n\n\n");
        if (S == NULL)
        {
            printf("\nno i am null\n");
            return;
        }
        printf("row is %ld column is %ld\n", S->row, S->column);
        printf("rowpointer:");
        for (int_t i = 0; i < S->row + 1; i++)
        {
            printf("%ld ", S->rowpointer[i]);
        }
        printf("\n");
        printf("columnindex:\n");
        for (int_t i = 0; i < S->row; i++)
        {
            for (int_t j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                printf("%ld ", S->columnindex[j]);
            }
            printf("\n");
        }
        printf("\n");
        printf("value:\n");
        for (int_t i = 0; i < S->row; i++)
        {
            for (int_t j = S->rowpointer[i]; j < S->rowpointer[i + 1]; j++)
            {
                printf("%lf ", S->value[j]);
            }
            printf("\n");
        }
        printf("\n\n\n--------------------");
    }
    void pangulu_memcpy_zero_pangulu_Smatrix_CSC_value(pangulu_Smatrix *S)
    {
        for(int_t i=0;i<S->nnz;i++){
            S->value_CSC[i]=0.0;
        }    
    }
    void pangulu_memcpy_zero_pangulu_Smatrix_CSR_value(pangulu_Smatrix *S)
    {
        for(int_t i=0;i<S->nnz;i++){
            S->value[i]=0.0;
        }    
    }
    void pangulu_display_pangulu_Smatrix_CSC(pangulu_Smatrix *S)
    {
        printf("------------------\n\n\n");
        if (S == NULL)
        {
            printf("\nno i am null\n");
            return;
        }
        printf("row is %ld column is %ld\n", S->row, S->column);
        printf("columnpointer:");
        for (int_t i = 0; i < S->row + 1; i++)
        {
            printf("%ld ", S->columnpointer[i]);
        }
        printf("\n");
        printf("rowindex:\n");
        for (int_t i = 0; i < S->row; i++)
        {
            for (int_t j = S->columnpointer[i]; j < S->columnpointer[i + 1]; j++)
            {
                printf("%ld ", S->rowindex[j]);
            }
            printf("\n");
        }
        printf("\n");
        printf("value_csc:\n");
        for (int_t i = 0; i < S->row; i++)
        {
            //printf("i %ld\n",i);
            for (int_t j = S->columnpointer[i]; j < S->columnpointer[i + 1]; j++)
            {
                printf("%lf ", S->value_CSC[j]);
            }
            printf("\n");
        }
        printf("\n\n\n--------------------");
    }

    void pangulu_display_pangulu_vector(pangulu_vector *V){
        printf("this vector length is %ld\n",V->row);
        for(int_t i=0;i<V->row;i++){
            printf("%lf\n",V->value[i]);
        }
    }

    double pangulu_get_spend_time(pangulu_common *common)
    {
        double time = (common->stop_time.tv_sec - common->start_time.tv_sec) * 1000.0 + (common->stop_time.tv_usec - common->start_time.tv_usec) / 1000.0;
        return time / 1000.0;
    }
    
    
    void pangulu_transport_pangulu_Smatrix_CSC_to_CSR(pangulu_Smatrix *S)
    {
#ifdef CHECK_TIME
        pangulu_time_check_begin();

#endif   
        // for (int_t i = 0; i < S->columnpointer[S->row]; i++)
        // {
        //     int_t index = S->CSR_to_CSC_index[i];
        //     S->value[i] = S->value_CSC[index];
        // }
        for (int_t i = 0; i < S->nnz; i++)
        {
            int_t index = S->CSC_to_CSR_index[i];
            S->value[index] = S->value_CSC[i];
        }
#ifdef CHECK_TIME
        TIME_transport+=pangulu_time_check_end();                              
#endif
    }
    void pangulu_transport_pangulu_Smatrix_CSR_to_CSC(pangulu_Smatrix *S)
    {

#ifdef CHECK_TIME
        pangulu_time_check_begin();

#endif
        // for (int_t i = 0; i < S->rowpointer[S->row]; i++)
        // {
        //     int_t index = S->CSR_to_CSC_index[i];
        //     S->value_CSC[index] = S->value[i];
        // }
        for (int_t i = 0; i < S->nnz; i++)
        {
            int_t index = S->CSC_to_CSR_index[i];
            S->value_CSC[i] = S->value[index];
        }

#ifdef CHECK_TIME
        TIME_transport+=pangulu_time_check_end();
#endif
    }

    void pangulu_pangulu_Smatrix_memcpy_rowpointer_CSR(pangulu_Smatrix *S,pangulu_Smatrix *copy_S)
    {
        int_t n=S->row;
        for(int_t i=0;i<(n+1);i++){
            S->rowpointer[i]=copy_S->rowpointer[i];
        }
    }

    void pangulu_pangulu_Smatrix_memcpy_value_CSR(pangulu_Smatrix *S,pangulu_Smatrix *copy_S)
    {
        for(int_t i=0;i<S->nnz;i++){
            S->value[i]=copy_S->value[i];
        }
    }

    void pangulu_pangulu_Smatrix_memcpy_value_CSR_copy_length(pangulu_Smatrix *S,pangulu_Smatrix *copy_S)
    {
        for(int_t i=0;i<copy_S->nnz;i++){
            S->value[i]=copy_S->value[i];
        }
    }
    
    void pangulu_pangulu_Smatrix_memcpy_struct_CSC(pangulu_Smatrix *S,pangulu_Smatrix *copy_S)
    {
        S->column=copy_S->column;
        S->row=copy_S->row;
        S->nnz=copy_S->nnz;
        for(int_t i=0;i<S->column+1;i++){
            S->columnpointer[i]=copy_S->columnpointer[i];
        }
        for(int_t i=0;i<S->nnz;i++){
            S->rowindex[i]=copy_S->rowindex[i];
        }
    }

    void pangulu_pangulu_Smatrix_memcpy_columnpointer_CSC(pangulu_Smatrix *S,pangulu_Smatrix *copy_S)
    {
        int_t n=S->column;
        for(int_t i=0;i<(n+1);i++){
            S->columnpointer[i]=copy_S->columnpointer[i];
        }
    }

    void pangulu_pangulu_Smatrix_memcpy_value_CSC(pangulu_Smatrix *S,pangulu_Smatrix *copy_S)
    {
        for(int_t i=0;i<S->nnz;i++){
            
            S->value_CSC[i]=copy_S->value_CSC[i];
        }
    }

    void pangulu_pangulu_Smatrix_memcpy_complete_CSC(pangulu_Smatrix *S,pangulu_Smatrix *copy_S)
    {
        pangulu_pangulu_Smatrix_memcpy_struct_CSC(S,copy_S);
        pangulu_pangulu_Smatrix_memcpy_value_CSC(S,copy_S);
    }

    void pangulu_pangulu_Smatrix_multiple_pangulu_vector_CSR(pangulu_Smatrix *A,
                                                             pangulu_vector *X,
                                                             pangulu_vector *B)
    {
        int_t n = A->row;
        calculate_type *X_value = X->value;
        calculate_type *B_value = B->value;
        for (int_t i = 0; i < n; i++)
        {
            for (int_t j = A->rowpointer[i]; j < A->rowpointer[i + 1]; j++)
            {
                int_t col = A->columnindex[j];
                B_value[i] += A->value[j] * X_value[col];
            }
        }
    }

    void pangulu_pangulu_Smatrix_multiply_block_pangulu_vector_CSR(pangulu_Smatrix *A,
                                                                   calculate_type *X,
                                                                   calculate_type *B)
    {
        int_t n = A->row;
        for (int_t i = 0; i < n; i++)
        {
            for (int_t j = A->rowpointer[i]; j < A->rowpointer[i + 1]; j++)
            {
                int_t col = A->columnindex[j];
                B[i] += A->value[j] * X[col];
            }
        }
    }

    void pangulu_pangulu_Smatrix_multiple_pangulu_vector_CSC(pangulu_Smatrix *A,
                                                             pangulu_vector *X,
                                                             pangulu_vector *B)
    {
        int_t n = A->column;
        calculate_type *X_value = X->value;
        calculate_type *B_value = B->value;
        for (int_t i = 0; i < n; i++)
        {
            for (int_t j = A->columnpointer[i]; j < A->columnpointer[i + 1]; j++)
            {
                int_t row = A->rowindex[j];
                B_value[row] += A->value_CSC[j] * X_value[i];
            }
        }
    }

    void pangulu_pangulu_Smatrix_multiply_block_pangulu_vector_CSC(pangulu_Smatrix *A,
                                                                   calculate_type *X,
                                                                   calculate_type *B)
    {
        int_t n = A->column;
        for (int_t i = 0; i < n; i++)
        {
            for (int_t j = A->columnpointer[i]; j < A->columnpointer[i + 1]; j++)
            {
                int_t row = A->rowindex[j];
                B[row] += A->value_CSC[j] * X[i];
            }
        }
    }

    void pangulu_read_pangulu_vector(pangulu_vector *X, int_t n)
    {
        X->value = (calculate_type *)pangulu_malloc(sizeof(calculate_type) * n);
        for (int_t i = 0; i < n; i++)
        {
            X->value[i] = (calculate_type)2.0;
        }
        X->row = n;
    }
    
    void pangulu_init_pangulu_vector(pangulu_vector *B, int_t n)
    {
        B->value = (calculate_type *)pangulu_malloc(sizeof(calculate_type) * n);
        for (int_t i = 0; i < n; i++)
        {
            B->value[i] = (calculate_type)0.0;
        }
        B->row = n;
    }

    void pangulu_send_pangulu_vector_value(pangulu_vector *S,
                                           int_t send_id, int_t signal, int_t vector_length)
    {
        MPI_Send(S->value, vector_length, MPI_VAL_TYPE, send_id, signal, MPI_COMM_WORLD);
    }

    void pangulu_recv_pangulu_vector_value(pangulu_vector *S, int_t receive_id, int_t signal, int_t vector_length)
    {
        MPI_Status status;
        MPI_Recv(S->value, vector_length, MPI_VAL_TYPE, receive_id, signal, MPI_COMM_WORLD, &status);
    }

    void pangulu_init_vector_int(int_t *vector,int_t length){
        for(int_t i=0;i<length;i++){
            vector[i]=0;
        }
    }

    int_t pangulu_choose_pivot(int_t i, int_t j)
    {
        return (i+j)/2;
    }

    void pangulu_swap_int( int_t *a ,  int_t *b)
    {
        int_t tmp = *a;
        *a = *b;
        *b = tmp;
    }

    void pangulu_quicksort_keyval(int_t *key, int_t *val, int_t start, int_t end)
    {
        int_t pivot;
        int_t i, j, k;

        if (start < end)
        {
            k = pangulu_choose_pivot(start, end);
            pangulu_swap_int(&key[start], &key[k]);
            pangulu_swap_int(&val[start], &val[k]);
            pivot = key[start];

            i = start + 1;
            j = end;
            while (i <= j)
            {
                while ((i <= end) && (key[i] <= pivot))
                    i++;
                while ((j >= start) && (key[j] > pivot))
                    j--;
                if (i < j)
                {
                    pangulu_swap_int(&key[i], &key[j]);
                    pangulu_swap_int(&val[i], &val[j]);
                }
            }

            // swap two elements
            pangulu_swap_int(&key[start], &key[j]);
            pangulu_swap_int(&val[start], &val[j]);
 
            // recursively sort the lesser key
            pangulu_quicksort_keyval(key, val, start, j-1);
            pangulu_quicksort_keyval(key, val, j+1, end);
        }
    }

    double pangulu_standard_deviation(int_t *p,int_t num){
        double average=0.0;
        for(int_t i=0;i<num;i++){
            average+=(double)p[i];
        }
        average/=(double)num;
        double answer=0.0;
        for(int_t i=0;i<num;i++){
            answer+=(double)(((double)p[i]-average)*((double)p[i]-average));
        }
        return sqrt(answer/(double)(num));

    }
    



#endif